
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : tmtable.scm
;; DESCRIPTION : tools for converting tables from and to other formats
;; COPYRIGHT   : (C) 2002  David Allouche, Joris van der Hoeven
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-module (convert tools tmtable))
(export tmtable-parser tmtable-block-borders tmtable-cell-halign)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Generic table parser
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmtable-parser x)
  ;; Internal state
  (let* ((contents			; table data
	  (map (lambda (row) (map cadr (cdr row)))
	       (do ((y x (last y)))
		   ((func? y 'table) (cdr y)))))
	 ;; table dimension
	 (nrcols (length (car contents)))
	 (nrrows (length contents))
	 ;; column-wide horizontal alignement
	 (cols-halign			; column-wide horizontal alignment 
	  (make-vector nrcols "l"))
	 ;; column wide borders
	 (cols-tborder (make-vector nrcols #f))
	 (cols-bborder (make-vector nrcols #f))
	 (cols-lborder (make-vector nrcols #f))
	 (cols-rborder (make-vector nrcols #f))
	 ;; row wide borders
	 (rows-tborder (make-vector nrrows #f))
	 (rows-bborder (make-vector nrrows #f))
	 (rows-lborder (make-vector nrrows #f))
	 (rows-rborder (make-vector nrrows #f)))


    ;; Public getters
    (define (global name)
      (cond ((eq? name 'border) (any-border?))
	    ((eq? name 'nrcols) nrcols)))
    (define (cols name)
      (cond ((eq? name 'halign) (vector->list cols-halign))
	    ((eq? name 'tborder) (vector->list cols-tborder))
	    ((eq? name 'bborder) (vector->list cols-bborder))
	    ((eq? name 'lborder) (vector->list cols-lborder))
	    ((eq? name 'rborder) (vector->list cols-rborder))))
    (define (rows name)
      (cond ((eq? name 'content) contents)
	    ((eq? name 'tborder) (vector->list rows-tborder))
	    ((eq? name 'bborder) (vector->list rows-bborder))
      	    ((eq? name 'lborder) (vector->list rows-lborder))
	    ((eq? name 'rborder) (vector->list rows-rborder))))


    ;; Public dispatcher
    (define (table-parser/dispatch scope . args)
      (apply (cond ((eq? scope 'global) global)
		   ((eq? scope 'cols) cols)
		   ((eq? scope 'rows) rows)) args))
  
    ;; Private
    (define (any-border?)
      (list-any noop
		(append (append-map cols '(tborder bborder lborder rborder))
			(append-map rows '(tborder bborder lborder rborder)))))
    
    (define (set-range! vect x1 x2 val)
      ;; Set a property for a range
      (do ((x x1 (1+ x)))
	  ((> x x2))
	(vector-set! vect x val)))
    
    (define (set-cols-halign! j1 j2 val)
      (set-range! cols-halign j1 j2 val))
    (define (set-cols-tborder! j1 j2 val)
      (set-range! cols-tborder j1 j2 val))
    (define (set-cols-bborder! j1 j2 val)
      (set-range! cols-bborder j1 j2 val))
    (define (set-cols-lborder! j1 j2 val)
      (set-range! cols-lborder j1 j2 val))
    (define (set-cols-rborder! j1 j2 val)
      (set-range! cols-rborder j1 j2 val))
    (define (set-rows-tborder! x1 x2 val)
      (set-range! rows-tborder x1 x2 val))
    (define (set-rows-bborder! x1 x2 val)
      (set-range! rows-bborder x1 x2 val))
    (define (set-rows-lborder! x1 x2 val)
      (set-range! rows-lborder x1 x2 val))
    (define (set-rows-rborder! x1 x2 val)
      (set-range! rows-rborder x1 x2 val))

    (define (parse-cwith x)
      (define (decode i n)
	(cond ((< i 0) (+ i n)) ((> i 0) (- i 1)) (else 0)))
      (let* ((I1 (string->number (second x)))
	     (I2 (string->number (third x)))
	     (J1 (string->number (fourth x)))
	     (J2 (string->number (fifth x)))
	     (i1 (decode I1 nrrows))
	     (i2 (decode I2 nrrows))
	     (j1 (decode J1 nrcols))
	     (j2 (decode J2 nrcols))
	     (var (sixth x))
	     (val (seventh x))
	     (range (cond ((== (list I1 I2 J1 J2) '(1 -1 1 -1)) 'table)
			  ((== (list I1 I2) '(1 -1)) 'col)
			  ((== (list J1 J2) '(1 -1)) 'row)
			  (else 'cell))))
	(values range i1 i2 j1 j2 var val)))

    (define (parse-tformat-item x)
      ;; Modify the internal state according to a table format entry. 
      (if (func? x 'cwith 6)
	  (receive (range i1 i2 j1 j2 var val) (parse-cwith x)		   
	    (let ((range>=col (in? range '(col table)))
		  (range>=row (in? range '(row table))))
	      (cond ((== var "cell halign")
		     (if (and range>=col (in? val '("l" "c" "r")))
			 (set-cols-halign! j1 j2 val)))
		    ((== var "cell lborder")
		     (if range>=col (set-cols-lborder! j1 j2 #t))
		     (if range>=row (set-rows-lborder! i1 i2 #t)))
		    ((== var "cell rborder")
		     (if range>=col (set-cols-rborder! j1 j2 #t))
		     (if range>=row (set-rows-rborder! i1 i2 #t)))
		    ((== var "cell tborder")
		     (if range>=col (set-cols-tborder! j1 j2 #t))
		     (if range>=row (set-rows-tborder! i1 i2 #t)))
		    ((== var "cell bborder")
		     (if range>=col (set-cols-bborder! j1 j2 #t))
		     (if range>=row (set-rows-bborder! i1 i2 #t))))))))

    ;; (trace-variables contents nrcols cols-halign global-border)
    ;; (set-trace-level! global cols rows)
    ;; (set-trace-level! any-border?)
    ;; (set-trace-level! set-cols-halign! set-cols-lborder! set-cols-rborder!)
    ;; (set-trace-level! set-rows-tborder! set-rows-bborder!)
    ;; (set-trace-level! parse-cwith parse-tformat-item)
    ;; (set-trace-level! table-parser/dispatch)

    (do ((x x (cAr x)))
	((func? x 'table))
      (for-each parse-tformat-item (cDdr x)))

    ;; evaluate to dispatcher closure
    table-parser/dispatch))

(define (tmtable-block-borders x)
  (if x '((cwith "1" "-1" "1" "1" "cell lborder" "1ln")
	  (cwith "1" "1" "1" "-1" "cell tborder" "1ln")
	  (cwith "1" "-1" "1" "-1" "cell bborder" "1ln")
	  (cwith "1" "-1" "1" "-1" "cell rborder" "1ln"))
      '()))

(define (tmtable-cell-halign x)
  `((cwith "1" "-1" "1" "-1" "cell halign" ,x)))
